use cpal::Sample;
use tracing::trace;

use super::sample::PcmSample;

const TAU: f64 = 2.0 * std::f64::consts::PI;

pub struct TwoStateCarrier {
    pub sample_rate: f64,
    pub symbol_rate: f64,
    pub freq_low: f64,
    pub freq_high: f64,
    phase: f64,
}

pub trait Carrier {
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample>;
    fn idle(&mut self) -> [PcmSample; 64];
}

impl Carrier for TwoStateCarrier {
    /// turn a u8 into bits & do FSK on the carrier for each, returning samples
    fn byte_into_fsk_samples(&mut self, byte: &u8) -> Vec<PcmSample> {
        //let mut bits: [u8; 8] = [0, 0, 0, 0, 0, 0, 0, 0];
        let mut out: Vec<PcmSample> = vec![];

        // LSb first
        //        bits[0] = (byte     ) & 1;
        //        bits[1] = (byte >> 1) & 1;
        //        bits[2] = (byte >> 2) & 1;
        //        bits[3] = (byte >> 3) & 1;
        //        bits[4] = (byte >> 4) & 1;
        //        bits[5] = (byte >> 5) & 1;
        //        bits[6] = (byte >> 6) & 1;
        //        bits[7] = (byte >> 7) & 1;

        // for each bit in the byte (LSb..=MSb)...
        for bit_place in 1..=8 {
            // if the bit is high...
            if (byte >> (bit_place - 1)) & 1 == 0 {
                // while the signal hasn't yet completed a single period...
                while self.phase < self.sample_rate / self.symbol_rate {
                    //trace!("generating lo sample for bit {bit_place}");
                    // push the next sample of the relative sine wave
                    out.push(
                        (self.phase * TAU * self.freq_low / self.sample_rate)
                            .sin(),
                    );
                    // next sample
                    self.phase += 1.0;
                }
                // if phase counter exceeds 1 period, drop it down again
            } else if (byte >> (bit_place - 1)) & 1 == 1 {
                while self.phase < self.sample_rate / self.symbol_rate {
                    //trace!("generating hi sample for bit {bit_place}");
                    out.push(
                        (self.phase * TAU * self.freq_high / self.sample_rate)
                            .sin(),
                    );
                    self.phase += 1.0;
                }
            } else {
                // this keeps happening
                unreachable!("FSK bit math returned a value that isn't 0 or 1");
            }
            self.phase = self.phase % (self.sample_rate / self.symbol_rate);
        }
        //trace!("FSK calculations turned up {:?}", &out);
        out
    }

    fn idle(&mut self) -> [PcmSample; 64] {
        let mut out: [PcmSample; 64] = [0.0; 64];
        for i in 0..=63 {
            out[i] =
                (self.phase * TAU * self.freq_low/ self.sample_rate).sin();
            self.phase += 1.0;
        }
        self.phase = self.phase % (self.sample_rate / self.freq_low);
        out
    }
}

impl TwoStateCarrier {
    pub fn new(freq_low: u32, freq_high: u32, sample_rate: u32, symbol_rate: u32) -> Self {
        TwoStateCarrier {
            freq_low: freq_low.into(),
            freq_high: freq_high.into(),
            sample_rate: sample_rate.into(),
            symbol_rate: symbol_rate.into(),
            phase: 0_f64,
        }
    }
    // setters
    fn freq_low(mut self, freq_low: u32) {
        self.freq_low = freq_low.into();
    }
    fn freq_high(mut self, freq_high: u32) {
        self.freq_high = freq_high.into();
    }
    fn sample_rate(mut self, sample_rate: u32) {
        self.sample_rate = sample_rate.into();
    }
    fn symbol_rate(mut self, symbol_rate: u32) {
        self.symbol_rate = symbol_rate.into();
    }
}
